/*
 * Copyright C.L.U.T. - Torino - Ottobre 2017
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part IV: Exam-Based Problems
 *
 * chapter 01 - exercise 05 - solution (2-1)
 *
 * Nessuna garanzia viene fornita su questo programma:
 * gli Autori e la Casa Editrice non potranno in nessun
 * caso essere ritenuti responsabili di eventuali danni
 * (materiali e/o immateriali) da esso cagionati.
 * E' liberamente consentita la modifica (ma non la
 * redistribuzione) di questo file, a patto che le note
 * di copyright poste in testa e in coda a esso vengano
 * mantenute inalterate.
 * 
 */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include "symbolTable.h"

#define MAX_LEN     (20+1)
#define MAX_ST_SIZE 1

// To debug the program set this to TRUE (not zero)
#define DEBUG       1

typedef struct sol_s {
  int *path;
  int *weight;
  int len;
  int sum;
} sol_t;

// Graph edge E
typedef struct edge_s {
  struct vertex_s *v;
  int weight;
  struct edge_s *next;
} edge_t;

typedef struct edge_t *edgesList;

// Graph vertex V
typedef struct vertex_s {
  int id, in_d, out_d;
  edge_t *el;
  struct vertex_s *next;
} vertex_t;

// Graph handle (wrapper)
typedef struct graph_s {
  vertex_t *head;
  vertex_t *tail;
  int el_n;
} graph_t;

// Prototypes
graph_t *graph_init ();
void graph_dispose (graph_t *);
vertex_t *vertex_init (int);
edge_t *edge_init (int, vertex_t *, int, edge_t *);
vertex_t* vertex_add (graph_t *, int);
void edge_add(graph_t *, int, int, int);
graph_t *file_read (char *, st_t *);
void graph_regular(graph_t *);
void path_find(graph_t *, int, st_t, sol_t *);
void path_find_r(graph_t *, vertex_t *, int, int *, sol_t *, sol_t *);
void graph_write(graph_t *, st_t);

int main(int argc, char **argv) {
  graph_t *graph = NULL;
  st_t vNames = NULL;
  char id[MAX_LEN];
  int i;
  sol_t best;

  if (argc != 2) {
    fprintf (stderr, "Run as: pgrmName inputFile\n");
    exit(EXIT_FAILURE);
  }

  graph = file_read(argv[1], &vNames);

#if DEBUG
  graph_write (graph, vNames);
#endif

  graph_regular(graph);

  best.len = best.sum = (-1);
  best.path = best.weight = NULL;
  while(1) {
    fprintf (stdout, "Insert vertex_t ID [end to quit] > ");
    fscanf(stdin, "%s", id);
    if (!strcmp(id, "end"))
      break;

    i = st_find(vNames, id);
    if (i == -1) {
      fprintf (stdout, "Invalid selection\n");
      continue;
    }
    path_find (graph, i, vNames, &best);
  }

  graph_dispose (graph);
  st_dispose (vNames);

  return 0;
}

// Allocate graph handle and init fields
graph_t *graph_init() {
  graph_t *g;

  g = (graph_t *) malloc(sizeof(graph_t));
  if (g == NULL) {
    fprintf (stderr, "Memory allocation error.\n");
    exit(EXIT_FAILURE);
  }
  g->head = NULL;
  g->tail = NULL;
  g->el_n = 0;

  return g;
}

// Free graph memory
void graph_dispose (graph_t *g) {
  vertex_t *v, *vtmp;
  edge_t *eli, *elitmp;

  vtmp = g->head;
  while(vtmp != NULL) {
    v = vtmp;
    vtmp = vtmp->next;

    elitmp = v->el;
    while(elitmp != NULL) {
      eli = elitmp;
      elitmp = elitmp->next;
      free(eli);
    }
    free(v);
  }
  free(g);
  return;
}

// Allocate and initialize a vertex
vertex_t *vertex_init(int id) {
  vertex_t *v = malloc(sizeof(vertex_t));

  if (v == NULL) {
    fprintf (stderr, "Memory allocation error.\n");
    exit(EXIT_FAILURE);
  }

  v->id = id;
  v->el = NULL;
  v->next = NULL;
  v->in_d = v->out_d = 0;

  return v;
}

// Allocate and initialize an edge
edge_t *edge_init(int dst, vertex_t *v, int w, edge_t * next) {
  edge_t *eli = malloc(sizeof(edge_t));

  if (eli == NULL) {
    fprintf (stderr, "Memory allocation error.\n");
    exit(EXIT_FAILURE);
  }

  eli->v = v;
  eli->weight = w;
  eli->next = next;

  return eli;
}

// Add a vertex to the graph
vertex_t* vertex_add(graph_t *g, int id) {
  if (g->head == NULL) {
    // First insertion...
    g->head = g->tail = vertex_init(id);
  } else {
    // Add in tail...
    g->tail->next = vertex_init(id);
    g->tail = g->tail->next;
  }
  g->el_n++;
  return g->tail;
}

// Add an edge to the graph
void edge_add(graph_t *g, int src, int dst, int weight) {
  if (g == NULL) {
    fprintf (stderr, "Memory allocation error.\n");
    exit(EXIT_FAILURE);
  }

  int i = 0, j = 0;
  vertex_t *t, *s, *d;
  for(t=g->head; t != NULL && i<src; t=t->next, i++);
  if (i == src && t != NULL) {
    // This vertex already appears in the head of the graph
    s = t;
  } else {
    s = vertex_add(g, src);
  }

  for(t=g->head; t != NULL && j<dst; t=t->next, j++);
  if (j == dst && t != NULL) {
    // This vertex already appears in the head of the graph
    d = t;
  } else {
    d = vertex_add(g, dst);
  }

  s->el = edge_init(dst, d, weight, s->el);
  s->out_d++;
  d->in_d++;
}

// Read input file
graph_t *file_read(char *name, st_t *vNames) {
  graph_t *graph;
  char v1[MAX_LEN], v2[MAX_LEN];
  int w, iv1, iv2;
  FILE *fp;

  fp = fopen (name, "r");
  if (fp == NULL) {
    fprintf (stderr, "File open error\n");
    exit(EXIT_FAILURE);
  }

  graph = graph_init();
  *vNames = st_init(MAX_ST_SIZE);

  while (fscanf(fp, "%s %d %s", v1, &w, v2) != EOF) {
    iv1 = st_find_or_add(*vNames, v1);
    iv2 = st_find_or_add(*vNames, v2);
    edge_add(graph, iv1, iv2, w);
  }

  fclose (fp);

  return (graph);
}

// Print-out graph
void graph_write(graph_t *g, st_t vNames) {
  vertex_t *v;
  edge_t *eli;
  int i = 0;

  fprintf (stdout, "GRAPH:\n");
  fprintf (stdout, "Num. of vertices: %d\n\n", g->el_n);
  for(v=g->head, i=0;v != NULL; v=v->next, i++) {
    fprintf (stdout, "#%d is %s [in:%d / out:%d]\n", i, st_get_index(vNames, i),
      v->in_d, v->out_d);
    for(eli=v->el;eli != NULL; eli=eli->next) {
      fprintf (stdout, "    goes to #%d [weight = %d]\n", eli->v->id, eli->weight);
    }
    fprintf (stdout, "\n");
  }
}

// Check regularity on a graph
void graph_regular(graph_t *g) {
  vertex_t *v;
  int in = -1, out = -1, reg = 1;

  for(v=g->head;v != NULL; v=v->next) {
    if (in == -1 && out == -1) {
      in = v->in_d;
      out = v->out_d;
      if (in != out) {
        reg = 0;
        break;
      }
    } else {
      if (in != v->in_d || out != v->out_d) {
        reg = 0;
        break;
      }
    }
  }

  if (reg) {
    fprintf (stdout, "The given graph is regular\n\n");
  } else {
    fprintf (stdout, "The given graph is NOT regular\n\n");
  }
}

// Find a path
void path_find (graph_t *g, int src, st_t vNames, sol_t *best) {
  sol_t curr;
  vertex_t *v = NULL;
  int i, *visited;

  visited = (int*) calloc (g->el_n, sizeof(int));
  curr.path = (int *) calloc (g->el_n, sizeof(int));
  curr.weight = (int *) calloc (g->el_n-1, sizeof(int));
  best->path = (int *) calloc (g->el_n, sizeof(int));
  best->weight = (int *) calloc (g->el_n-1, sizeof(int));
  if (visited==NULL || curr.weight==NULL || curr.path==NULL ||
      best->weight==NULL || best->path==NULL) {
    fprintf (stderr, "Memory allocation error.\n");
    exit(EXIT_FAILURE);
  }
  curr.sum = curr.len = 0;
  for(v=g->head, i=0; v != NULL && i<src; v=v->next, i++);

  path_find_r(g, v, src, visited, &curr, best);

  if (best->sum != -1) {
    for(i=0; i<=best->len; i++) {
      fprintf (stdout, "%s ", st_get_index(vNames, best->path[i]));
      if (i != best->len)
        fprintf (stdout, "-<w=%d>- ", best->weight[i]);
    }
    fprintf (stdout, ":: Wtot = %d\n", best->sum);
    best->sum = -1;
  } else {
    fprintf (stdout, "Nothing found...\n");
  }

  free(visited);
  free(curr.weight);
  free(curr.path);
}

// Find a path (recursive function)
void path_find_r(
  graph_t *g, vertex_t *v, int src, int *visited,
  sol_t *curr, sol_t *best)
{
  edge_t *eli;
  int i;

  if (visited[src] == 1) {
    return;
  }

  curr->path[curr->len] = src;
  if (best->sum == -1 || best->sum<curr->sum) {
    for (i=0; i<=curr->len; i++) {
      best->path[i] = curr->path[i];
      if (i != curr->len)
        best->weight[i] = curr->weight[i];
    }
    best->sum = curr->sum;
    best->len = curr->len;
  }

  visited[src] = 1;

  for(eli=v->el; eli!=NULL; eli=eli->next) {
    curr->weight[curr->len] = eli->weight;
    curr->len = curr->len + 1;
    curr->sum = curr->sum + eli->weight;
    path_find_r(g, eli->v, eli->v->id, visited, curr, best);
    curr->len = curr->len - 1;
    curr->sum = curr->sum - eli->weight;
  }

  visited[src] = 0;

  return;
}

/*
 * Copyright C.L.U.T. - Torino - Ottobre 2017
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part IV: Exam-Based Problems
 *
 */
