/*
 * Copyright C.L.U.T. - Torino - Ottobre 2017
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part IV: Exam-Based Problems
 *
 * chapter 01 - exercise 05 - solution (3-1)
 *
 * Nessuna garanzia viene fornita su questo programma:
 * gli Autori e la Casa Editrice non potranno in nessun
 * caso essere ritenuti responsabili di eventuali danni
 * (materiali e/o immateriali) da esso cagionati.
 * E' liberamente consentita la modifica (ma non la
 * redistribuzione) di questo file, a patto che le note
 * di copyright poste in testa e in coda a esso vengano
 * mantenute inalterate.
 * 
 */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include "graph.h"
#include "symbolTable.h"

#define MAX_LEN     (20+1)
#define MAX_ST_SIZE 1

// To debug the program set this to TRUE (not zero)
#define DEBUG       1

void edge_add(graph_t, int, int, int);
void file_read(graph_t *, char *, st_t *);
void graph_path_find (graph_t, int, st_t);

int main(int argc, char **argv) {
  graph_t graph = NULL;
  st_t vNames = NULL;
  char id[MAX_LEN];
  int i;

  if (argc != 2) {
    fprintf (stderr, "Run as: pgrm <file_name>\n");
    exit(EXIT_FAILURE);
  }
  file_read(&graph, argv[1], &vNames);

#if DEBUG
  graph_write(graph, vNames);
#endif

  if (graph_regular(graph)) {
    fprintf (stdout, "The graph is regular\n");
  } else {
    fprintf (stdout, "The graph is NOT regular\n");
  }

  while(1) {
    fprintf (stdout, "\nInsert vertex ID [end to quit] > ");
    fscanf (stdin, "%s", id);
    if (!strcmp(id, "end"))
      break;

    i = st_find(vNames, id);
    if (i == -1) {
      fprintf (stdout, "Invalid selection\n");
      continue;
    }
    graph_path_find (graph, i, vNames);
  }

  graph_dispose(graph);
  st_dispose(vNames);

  return 0;
}

void edge_add(graph_t g, int src, int dst, int weight) {
  if (g == NULL) {
    exit(EXIT_FAILURE);
  }
  graph_edge_insert(g, edge_init(src, dst, weight));
}

void file_read(graph_t *graph, char *name, st_t *vNames) {
  FILE *fp;
  char v1[MAX_LEN], v2[MAX_LEN];
  int w, iv1, iv2;

  *vNames = st_init(MAX_ST_SIZE);

  fp = fopen (name, "r");
  if (fp == NULL) {
    fprintf (stderr, "File open error.\n"); 
    exit(EXIT_FAILURE);
  }
  while(fscanf(fp, "%s %*d %s", v1, v2) != EOF) {
    iv1 = st_find_or_add(*vNames, v1);
    iv2 = st_find_or_add(*vNames, v2);
  }
  fclose (fp);

  fp = fopen (name, "r");
  *graph = graph_init(st_get_size(*vNames));
  while(fscanf(fp, "%s %d %s", v1, &w, v2) != EOF) {
    iv1 = st_find(*vNames, v1);
    iv2 = st_find(*vNames, v2);
    edge_add(*graph, iv1, iv2, w);
  }
  fclose (fp);
  
  return;
}

void graph_path_find (graph_t g, int src, st_t vNames) {
  int len = -1, w = 0, i;
  edge_t *path = path_find(g, src, &len, &w);

  if (path != NULL && len > 0) {
    for(i=0;i<len;i++)
      fprintf (stdout, "%s -- <w=%d> -- %s\n",
        st_get_index(vNames,path[i].v), path[i].wt, st_get_index(vNames,path[i].w));
    fprintf (stdout, "Total weight: %d\n", w);
  } else {
    fprintf (stdout, "No path found...\n");
  }

  free(path);
  return;
}


/*
 * Copyright C.L.U.T. - Torino - Ottobre 2017
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part IV: Exam-Based Problems
 *
 */
