/*
 * Copyright C.L.U.T. - Torino - Ottobre 2017
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part IV: Exam-Based Problems
 *
 * chapter 01 - exercise 05 - solution (3-1)
 *
 * Nessuna garanzia viene fornita su questo programma:
 * gli Autori e la Casa Editrice non potranno in nessun
 * caso essere ritenuti responsabili di eventuali danni
 * (materiali e/o immateriali) da esso cagionati.
 * E' liberamente consentita la modifica (ma non la
 * redistribuzione) di questo file, a patto che le note
 * di copyright poste in testa e in coda a esso vengano
 * mantenute inalterate.
 * 
 */

#include <stdlib.h>
#include <stdio.h>
#include <limits.h>
#include "graph.h"

#define maxWT 1000
#define maxV 100
#define DEBUG 0

typedef struct node_s node_t;
struct node_s {
  int v;
  int wt;
  node_t *next;
};

struct graph_s {
  int V;
  int E;
  node_t **adj;
};

static void path_find_r(
  graph_t, int, edge_t *, int *, int, int, edge_t *, int *, int *
);

node_t *node_init(int v, int wt, node_t *next) {
  node_t *x = malloc(sizeof *x);

  x->v = v;
  x->wt = wt;
  x->next = next;

  return x;
}

edge_t edge_init(int v, int w, int wt) {
  edge_t e;

  e.v = v;
  e.w = w;
  e.wt = wt;

  return e;
}

graph_t graph_init(int V) {
  int v;
  graph_t graph = malloc(sizeof *graph);

  graph->V = V;
  graph->E = 0;
  graph->adj = malloc( V  * sizeof(node_t *));

  for (v = 0; v < V; v++)
    graph->adj[v] = NULL;

  return graph;
}

void  graph_edge_insert(graph_t graph, edge_t e) {
  int v = e.v, w = e.w, wt = e.wt;

  if (v == w)
    return;

  graph->adj[v] = node_init(w, wt, graph->adj[v]);
  graph->E++;
}

int graph_edges(edge_t a[], graph_t graph) {
  int v, E = 0;
  node_t *t;

  for (v=0; v < graph->V; v++)
    for (t=graph->adj[v]; t != NULL; t = t->next)
      a[E++] = edge_init(v, t->v, t->wt);

  return E;
}

void graph_write(graph_t graph, st_t names) {
  int v;
  node_t *t;

  fprintf (stdout, "%d vertices, %d edges \n", graph->V, graph->E);
  if (names != NULL) {
    for (v=0; v < graph->V; v++) {
      fprintf (stdout, "%s:\t", st_get_index(names, v));
      for (t=graph->adj[v]; t != NULL; t = t->next) {
        fprintf (stdout, "%s/%d \t", st_get_index(names, t->v), t->wt);
      }
      fprintf (stdout, "\n");
    }
  } else {
    for (v=0; v < graph->V; v++) {
      fprintf (stdout, "%2d:", v);
      for (t=graph->adj[v]; t != NULL; t = t->next) {
        fprintf (stdout, "%d/%d \t", t->v, t->wt);
      }
      fprintf (stdout, "\n");
    }
  }
}

int graph_regular(graph_t graph) {
  int v;
  node_t *t;

  int *in_d = calloc(graph->V, sizeof(int));
  int *out_d = calloc(graph->V, sizeof(int));

  if (in_d==NULL || out_d==NULL) {
    fprintf (stderr, "Memory allocation error.\n");
    exit(EXIT_FAILURE);
  }

  for (v=0; v < graph->V; v++) {
    for (t=graph->adj[v]; t != NULL; t = t->next) {
      out_d[v]++;
      in_d[t->v]++;
    }
  }

  int reg = 1;
  for (v=0; v < graph->V; v++) {
    if ((in_d[v] != out_d[v]) || in_d[v] != in_d[0]) {
      reg = 0;
      break;
    }
  }

  free(in_d);
  free(out_d);

  return reg;
}

edge_t *path_find(graph_t graph, int src, int *len, int *weight) {
  edge_t *path = calloc(graph->E, sizeof(edge_t));
  edge_t *bestPath = calloc(graph->E, sizeof(edge_t));
  int *visited = calloc(graph->V, sizeof(int));

  if (path==NULL || bestPath==NULL || visited==NULL) {
    fprintf (stderr, "Memory allocation error.\n");
    exit(EXIT_FAILURE);
  }

  path_find_r(graph, src, path, visited, 0, 0, bestPath, len, weight);

  free(visited);
  free(path);

  return bestPath;
}

static void path_find_r(
  graph_t graph, int src, edge_t *path, int *visited, int curLen,
  int curW, edge_t *bestPath, int *bestLen, int *bestW
) {
  int i;
  node_t *t;

  if (visited[src]==1) {
    return;
  }

  visited[src] = 1;
  for (t=graph->adj[src]; t != NULL; t = t->next) {
    path[curLen] = edge_init(src, t->v, t->wt);
    path_find_r(graph, t->v, path, visited, curLen+1, curW+t->wt, bestPath,
      bestLen, bestW);
  }
  visited[src] = 0;

  // Update solution
  if (*bestW < curW || *bestLen == -1) {
    *bestW = curW;
    *bestLen = curLen;
    for(i=0;i<curLen;i++) bestPath[i] = path[i];
  }

  return;
}

void graph_dispose (graph_t graph) {
  int v;
  node_t *tmp, *t;

  for (v=0; v < graph->V; v++) {
    for (t=graph->adj[v]; t != NULL;) {
      tmp = t;
      t = t->next;
      free(tmp);
    }
  }
  free(graph->adj);
  free(graph);

  return;
}

/*
 * Copyright C.L.U.T. - Torino - Ottobre 2017
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part IV: Exam-Based Problems
 *
 */
