/*
 * Copyright C.L.U.T. - Torino - Ottobre 2017
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part IV: Exam-Based Problems
 *
 * chapter 01 - exercise 05 - solution (3-1)
 *
 * Nessuna garanzia viene fornita su questo programma:
 * gli Autori e la Casa Editrice non potranno in nessun
 * caso essere ritenuti responsabili di eventuali danni
 * (materiali e/o immateriali) da esso cagionati.
 * E' liberamente consentita la modifica (ma non la
 * redistribuzione) di questo file, a patto che le note
 * di copyright poste in testa e in coda a esso vengano
 * mantenute inalterate.
 * 
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "symbolTable.h"

struct st_s {
  char **array;
  int n_max;
  int n;
};

st_t st_init(int n_max) {
  st_t t;

  t = malloc(sizeof (struct st_s));
  if (t==NULL) {
    fprintf (stderr, "Memory allocation error.\n");
    exit(EXIT_FAILURE);
  }
  t->array = (char **) malloc(n_max * sizeof (char *));
  if (t->array==NULL) {
    fprintf (stderr, "Memory allocation error.\n");
    exit(EXIT_FAILURE);
  }

  t->n_max = n_max;
  t->n = 0;

  return t;
}

void st_dispose(st_t t) {
  int id;

  for (id=0; id<t->n; id++) {
    free(t->array[id]);
  }
  free(t->array);
  free(t);

  return;
}

int st_get_size(st_t t) {
  return t->n;
}

int st_add(st_t t, char *str) {
  int id = t->n;

  if (t->n >= t->n_max) {
    t->array = realloc(t->array, (2*t->n_max)*sizeof(char*));
    if (t->array == NULL) {
      fprintf (stderr, "Memory allocation error.\n");
      exit(EXIT_FAILURE);
    }
    t->n_max = 2*t->n_max;
  }
  t->array[id] = strdup (str);
  t->n++;

  return id;
}

int st_find(st_t t, char *str) {
  int id;

  for (id=0; id<t->n; id++) {
    if (strcmp(str,t->array[id])==0)
      return id;
  }
  return -1;
}

int st_find_or_add(st_t t, char *str) {
  int id = st_find(t,str);

  if (id>=0)
    return id;
  else
   return st_add(t,str);
}

char *st_get_index(st_t t, int id) {
  if (id < 0 || id>=t->n)
    return NULL;
  else
    return (t->array[id]);
}

/*
 * Copyright C.L.U.T. - Torino - Ottobre 2017
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part IV: Exam-Based Problems
 *
 */
