/*
 * Copyright C.L.U.T. - Torino - Ottobre 2017
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part IV: Exam-Based Problems
 *
 * chapter 01 - exercise 09 - solution (2-1)
 *
 * Nessuna garanzia viene fornita su questo programma:
 * gli Autori e la Casa Editrice non potranno in nessun
 * caso essere ritenuti responsabili di eventuali danni
 * (materiali e/o immateriali) da esso cagionati.
 * E' liberamente consentita la modifica (ma non la
 * redistribuzione) di questo file, a patto che le note
 * di copyright poste in testa e in coda a esso vengano
 * mantenute inalterate.
 * 
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define MAX (50+1)

typedef enum {M, F} sex;

// Main data structure
typedef struct node{
  char name[MAX];
  char surname[MAX];
  sex sex;
  int age;
  struct node *next;
  struct node *mother, *father, *children;
  struct node *nextBySex, *nextByAge;
} node_t;

// Prototypes
node_t* file_read(char *, int *);
node_t* person_search(node_t *, char *, char *, int);
node_t* parent_assign(node_t *, node_t *, node_t *);
node_t* child_assign(node_t *, node_t *);
int path_find_by_age(node_t *, int);
int path_find_by_sex(node_t *, int, sex);
void bySex_write(node_t *);
void byAge_write(node_t *);
node_t null_set(node_t);

int main (int argc, char *argv[]) {
  node_t *pt, *pathbySex, *pathbyAge;
  int n_people, i, j, max;
  
  if (argc!=2){
    fprintf (stderr, "Error in passing the arguments. Type <filename>");
    exit(EXIT_FAILURE);
  }
  
  pt = file_read (argv[1], &n_people);
  
  /* Path for Males */
  for (max=0, i=0; i<n_people; i++){
    if (pt[i].sex != M) continue;  // Also checked during recursion
    j = path_find_by_sex (&pt[i], 0, M);
    if (j>max) {
      max = j;
      pathbySex = &pt[i];
    }
  }
  fprintf (stdout, "Path by sex: %d males\n", max);
  bySex_write (pathbySex);

  /* Reset Links */
  for (i=0; i<n_people; i++){
    pt[i].nextBySex = NULL;
  }

  /* Path for Females */
  for (max=0, i=0; i<n_people; i++){
    if (pt[i].sex != F) continue;  // Also checked during recursion
    j = path_find_by_sex (&pt[i], 0, F);
    if (j>max) {
      max = j;
      pathbySex = &pt[i];
    }
  }
  fprintf (stdout, "Path by sex: %d females\n", max);
  bySex_write (pathbySex);

  /* Path for Longevity */
  for (i=0, max=0; i<n_people; i++) {
    j = path_find_by_age (&pt[i], 0);
    if (j>max) {
      max=j;
      pathbyAge = &pt[i];
    }
  }
  fprintf (stdout, "Path by age: %d years\n", max);
  byAge_write (pathbyAge);
  
  free(pt);
  
  return 0;
}

/*
 *  Read file and create array of ndde_t elements
 */
node_t* file_read (char* filename, int *n_people){
  FILE *fp;
  char sex, childN[MAX], childS[MAX];
  char momN[MAX], momS[MAX], dadN[MAX], dadS[MAX];
  int i=0;
  node_t *pt, *current, *curr_mom, *curr_dad;
  
  fp = fopen (filename, "r");
  if (fp==NULL) {
    fprintf (stderr, "File open error.\n");
    exit(EXIT_FAILURE);
  }
  
  if (fscanf (fp, "%d", n_people) == EOF) {
    fprintf (stderr, "Error in reading file.");
    exit(EXIT_FAILURE);
  }
  
  pt = (node_t *) malloc ((*n_people) * sizeof (node_t));
  if (pt==NULL){
    fprintf (stderr, "Error in allocating the memory.");
    exit(EXIT_FAILURE);
  }
  
  while (i<*n_people){
    fscanf(fp, "%s %s %c %d", pt[i].name, pt[i].surname, &sex, &pt[i].age);
    
    pt[i]=null_set(pt[i]);
    
    if (sex=='M') {
      pt[i].sex = M;
    } else {
      pt[i].sex = F;
    }
    
    i++;
  }
  
  while (fscanf (fp, "%s %s %s %s %s %s",
    childN, childS, dadN, dadS, momN, momS) != EOF) {
    current = person_search (pt, childN, childS, *n_people);
    curr_mom = person_search (pt, momN, momS, *n_people);
    curr_dad = person_search (pt, dadN, dadS, *n_people);
    current = parent_assign (current, curr_mom, curr_dad);
    curr_dad = child_assign (curr_dad, current);
    curr_mom = child_assign (curr_mom, current);
  }
  
  fclose(fp);
  
  return (pt);
}

/*
 *  Search a person within the array
 */
node_t *person_search (node_t* rp, char* name, char* surname, int n_people) {
  int i;
    
  for (i=0; i<n_people; i++){
    if (strcmp(rp[i].name, name)==0 &&
        strcmp(rp[i].surname, surname)==0) {
      return (&rp[i]);
    }
  }
    
  return NULL;
}

node_t null_set (node_t pt) {
  pt.mother = pt.father = pt.children = pt.next = NULL;
  pt.nextByAge = pt.nextBySex = NULL;
    
  return pt;
}

/*
 *  Set node pointers to given parents
 */
node_t *parent_assign (node_t* child, node_t* mom, node_t* dad) {
  child->mother = mom;
  child->father = dad;
    
  return (child);
}

/*
 *  Set a chld to a parent
 */
node_t *child_assign (node_t* parent, node_t* child) {
  child->next = parent->children;
  parent->children = child;
    
  return (parent);
}

void bySex_write (node_t* rp) {
  while (rp!=NULL) {
    fprintf (stdout, "%s %s\n", rp->name, rp->surname);
    rp=rp->nextBySex;
  }
    
  return;
}

void byAge_write(node_t* rp) {
  while (rp!=NULL) {
    fprintf (stdout, "%s %s\n", rp->name, rp->surname);
    rp = rp->nextByAge;
  }
  
  return;
}

/*
 *  Recursive search: Find the longest sequance of people with the same sex
 */ 
int path_find_by_sex (node_t* curr, int cnt, sex sex) {
  int i;
  node_t *pt=curr, *tmp = curr->children;
  
  if (pt==NULL || pt->sex!=sex)
    return (cnt);
  
  cnt++;
  
  while (pt->children != NULL){
    i = path_find_by_sex (pt->children, cnt, sex);

    if (i > cnt){
      cnt = i;
      (*curr).nextBySex = pt->children;
    }
   
    pt->children = pt->children->next;
  }
  
  pt->children = tmp;
  
  return (cnt);
}

/*
 *  Recursive search: Find the longest sequance of people with max lifespan
 */ 
int path_find_by_age (node_t* curr, int age) {
  int i, max=0;
  node_t *pt=curr, *tmp=curr->children;

  if (pt==NULL) return age;
  
  if (pt->children==NULL) return age+curr->age;
  
  while (pt->children != NULL) {
    i=path_find_by_age(pt->children, age+pt->age);
    if (i>max){
      max=i;
      (*curr).nextByAge = pt->children;
    }
      
    pt->children=pt->children->next;
  }
  
  pt->children=tmp;
  
  return max;
}

/*
 * Copyright C.L.U.T. - Torino - Ottobre 2017
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part IV: Exam-Based Problems
 *
 */
