/*
 * Copyright C.L.U.T. - Torino - Ottobre 2017
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part IV: Exam-Based Problems
 *
 * chapter 01 - exercise 13 - solution (2-1)
 *
 * Nessuna garanzia viene fornita su questo programma:
 * gli Autori e la Casa Editrice non potranno in nessun
 * caso essere ritenuti responsabili di eventuali danni
 * (materiali e/o immateriali) da esso cagionati.
 * E' liberamente consentita la modifica (ma non la
 * redistribuzione) di questo file, a patto che le note
 * di copyright poste in testa e in coda a esso vengano
 * mantenute inalterate.
 * 
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define DEBUG 1

typedef struct edge edge_t;
typedef struct node {
  char *name;
  int traversed;
  edge_t *edges;
  struct node *next;
} node_t;

typedef struct edge {
  int weight;
  struct edge *next;
  node_t *vertex;
} edge_t;

typedef struct graph {
  node_t *vertices;
} graph_t;

typedef struct path {
  char *name;
  struct path *next;
} path_t;

graph_t *file_read(char *);
void graph_write (graph_t *);
graph_t *graph_dispose (graph_t *);
node_t *search(node_t *, char *);
node_t *node_create(node_t **, char *);
edge_t *edge_create(edge_t *, node_t *, int);
void find_path(node_t *,char*, char *, int, int, char *);
int DFS_modified(node_t *, node_t *, path_t *, int, int, int, int, int,
  int, path_t **);
path_t *push(path_t *, char *);
path_t *pop(path_t *);
void list_visit(path_t *, FILE *);
path_t *list_copy(path_t *, path_t *);
path_t *list_dispose (path_t *);

int main(int argc, char *argv[]) {
  graph_t *g;
  char *name_in, *dep, *dest, *name_out;
  int k, p;

  if(argc!=7) {
    fprintf (stderr, "Run as: pgrm <in> <s> <d> <k> <p> <out>\n");
    exit(EXIT_FAILURE);
  }
  name_in = argv[1]; 
  dep = argv[2];
  k = atoi (argv[3]);
  p = atoi (argv[4]);
  dest = argv[5];
  name_out = argv[6];

  g = file_read (name_in);
#if DEBUG
  graph_write (g);
#endif
  find_path (g->vertices, dep, dest, k, p, name_out);
  graph_dispose (g);
  
  return 0;
}

/*
 *  Read graph from file
 */
graph_t *file_read (char *name) {
  FILE *fp;
  graph_t *g;
  char dep[20], dest[20];
  int weight;
  node_t* dep_p;
  node_t* dest_p;

  g = (graph_t*) malloc (1 * sizeof(graph_t));
  if (g==NULL) {
    fprintf (stderr, "Memory allocation error.\n");
    exit(EXIT_FAILURE);
  }
  g->vertices=NULL;

  fp = fopen (name, "r");
  if (fp == NULL) {
    fprintf (stderr, "File open error!\n");
    exit(EXIT_FAILURE);
  }

  while(fscanf(fp, "%s %d %s", dep, &weight, dest)!=EOF) {
    if((dep_p=search(g->vertices, dep)) == NULL) {
      dep_p=node_create(&(g->vertices), dep);
    }
    if((dest_p=search(g->vertices, dest))==NULL) {
      dest_p=node_create(&(g->vertices), dest);
    }
    dep_p->edges=edge_create(dep_p->edges, dest_p, weight);
  }

  fclose (fp);

  return g;
}

/*
 *  Print-out graph
 */
void graph_write (graph_t *g) {
  node_t *node_tmp;
  edge_t *edge_tmp;

  fprintf (stdout, "Graph:\n");
  node_tmp = g->vertices;
  while (node_tmp!=NULL) {
    fprintf (stdout, "%s -> ", node_tmp->name);
    edge_tmp = node_tmp->edges;
    while (edge_tmp!=NULL) {
      fprintf (stdout, "%s(%d) ", edge_tmp->vertex->name, edge_tmp->weight);
      edge_tmp = edge_tmp->next;
    }
    fprintf (stdout, "\n");
    node_tmp = node_tmp->next;
  }

  return;
}

/*
 *  Free graph
 */
graph_t *graph_dispose (graph_t *g) {
  node_t *node_tmp1, *node_tmp2;
  edge_t *edge_tmp1, *edge_tmp2;

  node_tmp1 = g->vertices;
  while (node_tmp1!=NULL) {
    node_tmp2 = node_tmp1;
    free (node_tmp1->name);
    edge_tmp1 = node_tmp1->edges;
    while (edge_tmp1!=NULL) {
      edge_tmp2 = edge_tmp1;
      free (edge_tmp1);
      edge_tmp1 = edge_tmp2->next;
    }
    node_tmp1 = node_tmp2->next;
  }
  free (g);

  return NULL;
}

/*
 *  Search a node within the node list
 */
node_t *search (node_t *head, char *value) {
  node_t *curr;

  curr = head;
  while (curr!=NULL) {
    if (strcmp(curr->name, value)==0) {
      return curr;
    }
    curr = curr->next;
  }

  return NULL;
}

/*
 *  Create a new node and assign fields
 */
node_t *node_create (node_t **head, char *value) {
  node_t *element;
  node_t *tmp;

  element = (node_t*) malloc (sizeof(node_t));
  if (element==NULL) {
    fprintf (stderr, "Memory allocation error.\n");
    exit(EXIT_FAILURE);
  }
  element->name = (char*) malloc ((strlen(value)+1)*sizeof(char));
  if (element->name==NULL) {
    fprintf (stderr, "Memory allocation error.\n");
    exit(EXIT_FAILURE);
  }
  strcpy(element->name, value);
  element->edges = NULL;
  element->traversed = 0;

  tmp = element;
  element->next = *head;
  *head = element;

  return tmp;
}

/*
 *  Create a new edge and assign fields
 */
edge_t *edge_create (edge_t* head_list, node_t* dest, int weight) {
  edge_t* element;

  element = (edge_t*) malloc (sizeof(edge_t));
  if (element==NULL) {
    fprintf (stderr, "Memory allocation error.\n");
    exit(EXIT_FAILURE);
  }
  element->weight=weight;
  element->vertex=dest;
  element->next=head_list;
  head_list=element;

  return head_list;
}

/*
 *  Set up everything to run modified DFS
 */
void find_path (node_t* vertices, char *dep, char *dest, int k, int p, char *name) {
  node_t *dep_p;
  node_t *dest_p;
  path_t *path_current;
  path_t *path_best;
  FILE *fp;
  int max;

  max = 0;
  path_best = NULL;
  path_current = NULL;

  dep_p = search (vertices, dep);
  dest_p = search (vertices, dest);

  max = DFS_modified(dep_p, dest_p, path_current, 0, 0, k, p, 0, max, &path_best);
  fprintf (stdout, "Satisfying path: ");
  list_visit (path_best, stdout);
  fprintf (stdout, "Total weight: %d\n", max);

  fp = fopen(name, "w");
  if (fp == NULL) {
    exit(EXIT_FAILURE);
  }
  list_visit (path_best, fp);
  list_dispose (path_best);
  fclose (fp);

  return;
}

/*
 *  Modified Depth First Search (DFS) routine
 */
int DFS_modified (
  node_t *dep, node_t *dest, path_t *path_current, int k, int p, int k_max,
  int p_max, int sum_current, int sum_max, path_t **path_best)
{
  edge_t *loc_edge;
  loc_edge = dep->edges;

  if (dep==dest) {
    if (sum_current>sum_max) {
      sum_max = sum_current;
      *path_best = list_dispose (*path_best);
      *path_best = push (*path_best, dest->name);
      *path_best = list_copy (*path_best, path_current);
    }
    return sum_max;
  }

  if(dep->traversed>0) {
    if (k>k_max) {
      return sum_max;
    } else {
      if(p+1>p_max) {
	return sum_max;
      }
    }
    k++;
    p++;
  }

  dep->traversed++;
  path_current = push (path_current, dep->name);
  while (loc_edge!=NULL) {
    sum_max = DFS_modified(loc_edge->vertex, dest, path_current, k, p, k_max,
              p_max,sum_current+loc_edge->weight, sum_max, path_best);
    loc_edge = loc_edge->next;
  }
  path_current = pop(path_current);
  dep->traversed--;

  return sum_max;
}

/*
 *  Library push into a LIFO stack data structure
 */
path_t *push (path_t *head, char *data) {
  path_t *element;
  
  element = (path_t *) malloc (sizeof(path_t));
  if (element==NULL) {
    fprintf (stderr, "Memory allocation error.\n");
    exit(EXIT_FAILURE);
  }
  element->name = (char *) malloc ((strlen(data)+1) * sizeof(char));
  if (element->name==NULL) {
    fprintf (stderr, "Memory allocation error.\n");
    exit(EXIT_FAILURE);
  }
  strcpy (element->name, data);
  element->next = head;
  head = element;
  
  return head;
}

/*
 *  Library pop from a LIFO stack data structure
 */
path_t *pop (path_t *head) {
  path_t *old;

  if (head==NULL) {
    fprintf (stderr, "Pop error!\n");
  } else {
    old = head;
    head = head->next;
    free(old);
  }
  
  return head;
}

/*
 *  Library list visit
 */
void list_visit (path_t *head, FILE *fp) {
  path_t* curr;
  
  curr = head;
  while(curr!=NULL) {
    fprintf (fp, "%s ", curr->name);
    curr=curr->next;
  }
  fprintf (fp, "\n");

  return;
}

/*
 *  Library list copy (duplication)
 */
path_t *list_copy (path_t *path_best, path_t *path_current) {
  path_t* curr;
  
  curr = path_current;
  while (curr!=NULL) {
    path_best = push (path_best, curr->name);
    curr = curr->next;
  }
  
  return path_best;
}

/*
 *  Library list free
 */
path_t *list_dispose (path_t *path) {
  path_t *curr, *next;
  
  curr = path;
  while (curr!=NULL) {
    next = curr->next,
    free (curr->name);
    free (curr);
    curr = next;
  }
  
  return NULL;
}


/*
 * Copyright C.L.U.T. - Torino - Ottobre 2017
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part IV: Exam-Based Problems
 *
 */
