/*
 * Copyright C.L.U.T. - Torino - Ottobre 2017
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part IV: Exam-Based Problems
 *
 * chapter 01 - exercise 15 - solution (2-1)
 *
 * Nessuna garanzia viene fornita su questo programma:
 * gli Autori e la Casa Editrice non potranno in nessun
 * caso essere ritenuti responsabili di eventuali danni
 * (materiali e/o immateriali) da esso cagionati.
 * E' liberamente consentita la modifica (ma non la
 * redistribuzione) di questo file, a patto che le note
 * di copyright poste in testa e in coda a esso vengano
 * mantenute inalterate.
 * 
 */

#include <stdio.h>
#include <stdlib.h>

typedef struct vertex {
  int id;
  int flag;
  int n_edge;
  struct edge *edges;
  struct vertex *next;
} vertex_t;

typedef struct edge {
  vertex_t *vertex;
} edge_t;

typedef struct graph {
  int n_vertices;
  vertex_t *vertices;
} graph_t;

graph_t *file_read(char *, int *);
vertex_t *vertex_search(graph_t *, int);
int recur(graph_t *, int, int);
int check(graph_t *);
void bfs(vertex_t *, int);
void *data_alloc(int thesize);
void data_dispose(graph_t *);

int main (int argc, char *argv[]) {
  graph_t *g;
  int i,n;
  vertex_t *p;

  if (argc!=2) {
    fprintf (stderr, "Argument error.\n");
    exit(EXIT_FAILURE);
  }
  g = file_read (argv[1],&n);

  for(i=1; i<=n; i++) {
    if(recur(g, 0, i)) {
      fprintf (stdout, "Minimum number of edges to rule-out is %d\n", i);
      break;
    }
  }
  for (p=g->vertices; p!=NULL; p=p->next) {
    for (i=0; i<p->n_edge; i++)
      if (p->edges[i].vertex==NULL)
        fprintf (stdout, "Terminal node of the edge which ruled-out is: %d\n",
          p->id+1);
  }

  data_dispose (g);

  return 0;
}

/*
 *  Read input file
 */
graph_t *file_read(char *fname, int *n) {
  graph_t *g;
  vertex_t *p, *q;
  FILE *fp;
  int i,j,k;

  g = NULL;
  fp = fopen (fname, "r");
  if(fp==NULL) {
    fprintf (stderr, "File open error.\n");
    exit(EXIT_FAILURE);
  }
  if(fscanf(fp,"%d",n) != 1) {
    fprintf (stderr, "Error on file format.\n");
    exit(EXIT_FAILURE);
  }
  g = (graph_t *) data_alloc (sizeof(graph_t));
  g->n_vertices = *n;
  g->vertices = NULL;

  // For each row ...
  for (i=0; i<g->n_vertices; i++) {
    p = vertex_search (g, i);
    if(p==NULL) {
      p = (vertex_t *) data_alloc (sizeof(vertex_t));
      p->id = i;
      p->flag = 0;
      p->n_edge = 0;
      p->edges = NULL;
      p->next = g->vertices;
      g->vertices = p;
    }
    // ... read entire row
    for (j=0; j<g->n_vertices; j++) {
      fscanf (fp, "%d", &k);
      if(k==1) {
        q = vertex_search (g, j);
        if (q==NULL) {
          q = (vertex_t *) data_alloc (sizeof(vertex_t));
          q->id = j;
          q->flag = 0;
          q->n_edge = 0;
          q->edges = NULL;
          q->next = g->vertices;
          g->vertices = q;
        }
        p->n_edge++;
        p->edges = realloc (p->edges, p->n_edge*sizeof(edge_t));
        p->edges[p->n_edge-1].vertex=q;
      }
    }
  }
  fclose(fp);

  return(g);
}

vertex_t *vertex_search (graph_t *g, int theid) {
  vertex_t *p;

  if(g==NULL) {
    return NULL;
  }
  p = g->vertices;
  while (p!=NULL && p->id!=theid) {
    p = p->next;
  }
  if (p!=NULL) {
    return (p);
  } else {
    return NULL;
  }
}

/*
 *  Recursively delete num edges and check for connectivity
 */
int recur (graph_t *g, int step, int num) {
  vertex_t *p, *q;
  int i,j;

  if(step==num) {
    if(check(g))
      return (1);
    else
      return (0);
  }

  p = g->vertices;
  while (p!=NULL) {
    for (i=0;i<p->n_edge;i++) {
      if (p->edges[i].vertex!=NULL) {
        q = p->edges[i].vertex;
        for(j=0;j<q->n_edge;j++) {
          if(q->edges[j].vertex!=NULL && q->edges[j].vertex->id==p->id) {
            break;
          }
        }
        p->edges[i].vertex=NULL;
        q->edges[j].vertex=NULL;
        if(recur(g,step+1,num)) {
          return (1);
        }
        p->edges[i].vertex=q;
        q->edges[j].vertex=p;
      }
    }
    p=p->next;
  }

  return (0);
}

/*
 *  Check connectivity
 */
int check (graph_t *g) {
  vertex_t *p;

  p = g->vertices;
  bfs (p, 1);
  p = g->vertices;
  while (p!=NULL) {
    if (p->flag==0) {
      return (1);
    }
    p = p->next;
  }
  p = g->vertices;
  bfs (p, 0);

  return 0;
}

/*
 *  Breadth First Search (BFS)
 */
void bfs (vertex_t *p, int k) {
  int i;

  p->flag = k;
  for(i=0; i<p->n_edge; i++) {
    if(p->edges[i].vertex!=NULL && p->edges[i].vertex->flag!=k)
      bfs (p->edges[i].vertex, k);
  }

  return;
}

/*
 *  Free data structure
 */
void data_dispose(graph_t *g) {
  vertex_t *p, *q;

  p = g->vertices;
  while(p!=NULL) {
    q = p;
    p = p->next;
    free (q->edges);
    free (q);
  }
  free (g);

  return;
}

/*
 *  Memory allocation
 */
void *data_alloc (int thesize) {
  void *temp;

  temp = (void *) malloc( thesize);
  if (temp==NULL) {
    fprintf (stderr, "Memory allocation error.\n");
    exit(EXIT_FAILURE);
  }

  return (temp);
}

/*
 * Copyright C.L.U.T. - Torino - Ottobre 2017
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part IV: Exam-Based Problems
 *
 */
