/*
 * Copyright C.L.U.T. - Torino - Ottobre 2017
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part IV: Exam-Based Problems
 *
 * chapter 01 - exercise 17 - solution (2-1)
 *
 * Nessuna garanzia viene fornita su questo programma:
 * gli Autori e la Casa Editrice non potranno in nessun
 * caso essere ritenuti responsabili di eventuali danni
 * (materiali e/o immateriali) da esso cagionati.
 * E' liberamente consentita la modifica (ma non la
 * redistribuzione) di questo file, a patto che le note
 * di copyright poste in testa e in coda a esso vengano
 * mantenute inalterate.
 * 
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

// Set to TRUE to print some debug message
#define DEBUG 0

int play (int **, int **, int **, int, int, int);
int check (int **, int **, int **, int, int);
void constraint_write (int **, int);
void mat_write (FILE *, int **, int, int);

int main (int argc, char *argv[]) {
  FILE *fpR, *fpW;
  int **matR, **matC, **nonogram;
  int i, j, n, nR, nC;

  fpR = fopen (argv[1], "r");
  fpW = fopen (argv[2], "w");
  if (fpR==NULL || fpW==NULL) {
    fprintf (stdout, "File open error.\n");
    exit(EXIT_FAILURE);
  }

  // Read constraints for all nonogram rows
  fscanf (fpR, "%d", &nR);
  matR = (int **) malloc (nR * sizeof (int *));
  if (matR==NULL) {
    fprintf (stderr, "Memory allocation error.\n");
    exit(EXIT_FAILURE);
  }
  for (i=0; i<nR; i++) {
    fscanf (fpR, "%d", &n);
    matR[i] = (int *) malloc ((n+1) * sizeof (int));
    if (matR[i]==NULL) {
      fprintf (stderr, "Memory allocation error.\n");
      exit(EXIT_FAILURE);
    }
    matR[i][0] = n;
    for (j=1; j<=n; j++) {
      fscanf (fpR, "%d", &matR[i][j]);
    }
  }

  // Read constraints for all nonogram columns
  fscanf (fpR, "%d", &nC);
  matC = (int **) malloc (nC * sizeof (int *));
  if (matC==NULL) {
    fprintf (stderr, "Memory allocation error.\n");
    exit(EXIT_FAILURE);
  }
  for (i=0; i<nC; i++) {
    fscanf (fpR, "%d", &n);
    matC[i] = (int *) malloc ((n+1) * sizeof (int));
    if (matC==NULL) {
      fprintf (stderr, "Memory allocation error.\n");
      exit(EXIT_FAILURE);
    }
    matC[i][0] = n;
    for (j=1; j<=n; j++) {
      fscanf (fpR, "%d", &matC[i][j]);
    }
  }

  fclose (fpR);

#if DEBUG
  constraint_write (matR, nR);
  constraint_write (matC, nC);
#endif

  // Allocate and initialize nonogram matrix
  nonogram = (int **) malloc (nR * sizeof (int *));
  if (nonogram==NULL) {
    fprintf (stderr, "Memory allocation error.\n");
    exit(EXIT_FAILURE);
  }
  for (i=0; i<nR; i++) {
    nonogram[i] = (int *) malloc (nC * sizeof (int));
    if (nonogram==NULL) {
      fprintf (stderr, "Memory allocation error.\n");
      exit(EXIT_FAILURE);
    }
  }

  for (i=0; i<nR; i++) {
    for (j=0; j<nC; j++) {
      nonogram[i][j] = 0;
    }
  }

  // Find solution
  if (play (nonogram, matR, matC, nR, nC, 0) == 1) {
    fprintf (stdout, "Found Solution:\n");
    mat_write (stdout, nonogram, nR, nC);
    mat_write (fpW, nonogram, nR, nC);
  }

  // Free all memory
  for (i=0; i<nR; i++) {
    free (matR[i]);
  }
  free (matR);
  for (i=0; i<nC; i++) {
    free (matC[i]);
  }
  free (matC);
  for (i=0; i<nR; i++) {
    free (nonogram[i]);
  }
  free (nonogram);

  fclose (fpW);

  return (0);
}

/*
 *  Check reading phase (write constrints)
 */
void constraint_write (int **mat, int n) {
  int r, c;

  fprintf (stdout, "%d\n", n);
  for (r=0; r<n; r++) {
    for (c=0; c<=mat[r][0]; c++) {
      fprintf (stdout, "%d ", mat[r][c]);
    }
    fprintf (stdout, "\n");
  }
  
  return;
}

/*
 *  Print solution
 */
void mat_write (FILE *fpW, int **mat, int nR, int nC) {
  int r, c;

  for (r=0; r<nR; r++) {
    for (c=0; c<nC; c++) {
      fprintf (fpW, "%d ", mat[r][c]);
    }
    fprintf (fpW, "\n");
  }
  
  return;
}

/*
 *  Find solution
 */
int play (int **nonogram, int **matR, int **matC, int nR, int nC, int level) {
  int r, c;

  if (level == (nR*nC)) {
    return (check (nonogram, matR, matC, nR, nC));
  }

  /* From recursion level to matrix indexes */
  r = (int) (level / nC);
  c = (int) (level % nC);

  /* Try all possible combinations of white/clack cells */
  nonogram[r][c] = 0;
  if (play (nonogram, matR, matC, nR, nC, level+1) == 1) {
    return (1);
  }
  nonogram[r][c] = 1;
  if (play (nonogram, matR, matC, nR, nC, level+1) == 1) {
    return (1);
  }
  
  return (0);
}

/*
 *  Check current solution against row and column constraints
 */
int check (int **nonogram, int **matR, int **matC, int nR, int nC) {
  int r, c, n, l;

#if DEBUG
  static int checkN = 0;

  fprintf (stdout, "NONOGRAMS %d\n", checkN++);

  for (r=0; r<nR; r++) {
    for (c=0; c<nC; c++) {
      fprintf (stdout, "%d ", nonogram[r][c]);
    }
    fprintf (stdout, "\n");
  }
#endif

  /* Row Check */
  for (r=0;r<nR; r++) {
    n = 0;
    l = 0;

    for (c=0; c<nC; c++) {
      if (nonogram[r][c]==1) {
        l++; if (l==1)n++;
      } else {
        if (l!=0) {
          if (n>matR[r][0] || l!=matR[r][n]) {
            return (0);
          } else {
            l = 0;
          }
        }
      }
    }

    if (n!=matR[r][0])
      return (0);
    if (l!=0 && l!=matR[r][n])
      return (0);
  }

  /* Column Check */
  for (c=0; c<nC; c++) {
    n = 0;
    l = 0;

    for (r=0; r<nR; r++) {
      if (nonogram[r][c]==1) {
        l++; if (l==1)n++;
      } else {
        if (l!=0) {
          if (n>matC[c][0] || l!=matC[c][n]) {
            return (0);
          } else {
            l = 0;
          }
        }
      }
    }

    if (n!=matC[c][0])
      return (0);
    if (l!=0 && l!=matC[c][n])
      return (0);
  }

  return (1);
}

/*
 * Copyright C.L.U.T. - Torino - Ottobre 2017
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part IV: Exam-Based Problems
 *
 */
