/*
 * Copyright C.L.U.T. - Torino - Ottobre 2017
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part IV: Exam-Based Problems
 *
 * chapter 01 - exercise 19 - solution (2-1)
 *
 * Nessuna garanzia viene fornita su questo programma:
 * gli Autori e la Casa Editrice non potranno in nessun
 * caso essere ritenuti responsabili di eventuali danni
 * (materiali e/o immateriali) da esso cagionati.
 * E' liberamente consentita la modifica (ma non la
 * redistribuzione) di questo file, a patto che le note
 * di copyright poste in testa e in coda a esso vengano
 * mantenute inalterate.
 * 
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <limits.h>

#define L           21
#define NO_MINISTRY  0
#define MINISTRY     1

#define LINEAR       1

/* City element */
struct e {
  char *name;
  int citizen;
  int distance;
  int status;
  int best;
  };

/* Structure wrapper */
struct handle {
  struct e *city;
  int n;
  int k;
  int costBest;  
  };

void file_read (char *, struct handle *);
void ministry_place (struct handle *, int, int);
#if LINEAR
  void costLinear (struct handle *);
#else
  void costQuadratic (struct handle *);
#endif
void solution_write (struct handle);
void myFree (struct handle *);

int main (int argc, char *argv[]) {
  char name[L];
  struct handle ds;

  if (argc!=3) {
    fprintf (stderr, "Run program as: pgrm fileName k\n");
    exit(EXIT_FAILURE);
  }

  strcpy (name, argv[1]);
  ds.k = atoi (argv[2]);
  ds.costBest = INT_MAX;

  file_read (name, &ds);

  ministry_place (&ds, 0, 0);
  solution_write (ds);

  myFree (&ds);

  return (1);
}

/*
 *  Read the input file
 */
void file_read (char *name, struct handle *ds) {
  FILE *fp;
  int i, c, d, n;
  struct e *city;

  fp = fopen (name, "r");
  if (fp==NULL) {
    fprintf (stderr, "File open error.\n");
    exit(EXIT_FAILURE);
  }

  if (fscanf (fp, "%d", &n) == EOF) {
    fprintf (stderr, "Read file error.\n");
    exit(EXIT_FAILURE);
  }

  city = (struct e *) malloc (n * sizeof (struct e));
  if (city==NULL) {
    fprintf (stderr, "Memory allocation error.\n");
    exit(EXIT_FAILURE);
  }
  
  i = 0;
  while (fscanf (fp, "%s%d%d", name, &c, &d) != EOF || i<n) {
    city[i].name = strdup (name);
    city[i].citizen = c;
    city[i].distance = d;
    city[i].status = NO_MINISTRY;
    city[i].best = NO_MINISTRY;
    i++;
  }

  ds->n = n;
  ds->city = city;
  fclose (fp);

  return;
}

/*
 *  Place ministries recursively
 */
void ministry_place (struct handle *ds, int from, int level) {
  int i;

  if (level==ds->k) {
#if LINEAR
    costLinear (ds);
#else
    costQuadratic (ds);
#endif
    return;
  }
  
  for (i=from; i<ds->n; i++) {
    if (ds->city[i].status == NO_MINISTRY) {
      ds->city[i].status = MINISTRY;
      ministry_place (ds, i+1, level+1);
      ds->city[i].status = NO_MINISTRY;
    }
  }

  return;
}

/*
 *  Compute the cost: version 1
 */
void costQuadratic (struct handle *ds) {
  int i, j, dLeft, dRight, dMin, cost;

  cost = 0;
  /* For each city */
  for (i=0; i<ds->n; i++) {
    dLeft = dRight = INT_MAX;
    /* Compute distance on the left */
    for (j=i; j>=0; j--) {
      if (ds->city[j].status==MINISTRY) {
        dLeft = ds->city[i].distance - ds->city[j].distance;
        break;
      }
    }
    /* Compute distance on the right */
    for (j=i+1; j<ds->n; j++) {
      if (ds->city[j].status==MINISTRY) {
        dRight = ds->city[j].distance - ds->city[i].distance;
        break;
      }
    }
    /* Compute cost */
    dMin = (dLeft<dRight) ? dLeft :dRight,
    cost = cost + ds->city[i].citizen * dMin;
  }

  /* Compare current solution with the best one */
  if (cost < ds->costBest) {
    ds->costBest = cost;
    for (i=0; i<ds->n; i++) {
      ds->city[i].best = ds->city[i].status;
      ds->city[i].status = NO_MINISTRY;
    }
  }

  return;
}

/*
 *  Compute the cost: version 2
 */
void costLinear (struct handle *ds) {
  int i, d, cost, tmp;
  int *dArray;

  dArray = (int *) malloc (ds->n * sizeof (int));
  if (dArray==NULL) {
    fprintf (stderr, "Memory allocation error.\n");
    exit(EXIT_FAILURE);
  }
  
  /* Proceed from left to right and
     compute distance from the closest ministry */
  d = (-1);
  for (i=0; i<ds->n; i++) {
    if (ds->city[i].status==MINISTRY) {
      dArray[i] = 0;
      d = ds->city[i].distance;
    } else {
      if (d==(-1)) {
        dArray[i] = INT_MAX;
      } else {
        dArray[i] = ds->city[i].distance - d;
      }
    }
  }

  /* Proceed from right to left and
     compute distance from the closest ministry
     included the one evaluated from left to right */
  d = (-1);
  for (i=ds->n-1; i>=0; i--) {
    if (ds->city[i].status==MINISTRY) {
      d = ds->city[i].distance;
    } else {
      if (d!=(-1)) {
        tmp = d - (ds->city[i].distance);
        if (tmp<dArray[i]) {
          dArray[i] = tmp;
        }
      }
    }
  }

  /* Compute final cost */
  cost = 0;
  for (i=0; i<ds->n; i++) {
    cost = cost + ds->city[i].citizen * dArray[i];
  }

  /* Compare current solution with the best one */
  if (cost < ds->costBest) {
    ds->costBest = cost;
    for (i=0; i<ds->n; i++) {
      ds->city[i].best = ds->city[i].status;
      ds->city[i].status = NO_MINISTRY;
    }
  }

  free (dArray);

  return;
}

/*
 *  Print solution
 */
void solution_write (struct handle ds) {
  int i;

  fprintf (stdout, "Ministry placed in:\n");
  for (i=0; i<ds.n; i++) {
    if (ds.city[i].best == MINISTRY) {
      fprintf (stdout, "#%d City=%s Citizen=%d Distance=%d\n",
        i, ds.city[i].name, ds.city[i].citizen, ds.city[i].distance);
    }
  }
  fprintf (stdout, "Cost=%d\n", ds.costBest);

  return;
}

void myFree (struct handle *ds) {
  int i;
  struct e *city;

  city = ds->city;
  for (i=0; i<ds->n; i++) {
    free (city[i].name);
  }
  free (city);
  ds->city = NULL;

  return;
}


/*
 * Copyright C.L.U.T. - Torino - Ottobre 2017
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part IV: Exam-Based Problems
 *
 */
