/*
 * Copyright C.L.U.T. - Torino - Ottobre 2017
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part IV: Exam-Based Problems
 *
 * chapter 01 - exercise 21 - solution (2-1)
 *
 * Nessuna garanzia viene fornita su questo programma:
 * gli Autori e la Casa Editrice non potranno in nessun
 * caso essere ritenuti responsabili di eventuali danni
 * (materiali e/o immateriali) da esso cagionati.
 * E' liberamente consentita la modifica (ma non la
 * redistribuzione) di questo file, a patto che le note
 * di copyright poste in testa e in coda a esso vengano
 * mantenute inalterate.
 * 
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "bst.h"

typedef struct bst_node_s {
  char *key;
  int index;
  struct bst_node_s *l;
  struct bst_node_s *r;
} bst_node_t;

struct bst_s {
  int N;
  bst_node_t *head;
  bst_node_t *z;
};

bst_node_t *tree_node_new(char *key, int index, bst_node_t *l, bst_node_t *r) {
  bst_node_t *x;

  x = malloc (sizeof (bst_node_t));
  if (x == NULL) {
    fprintf(stderr, "Memory allocation error.\n");
    exit(EXIT_FAILURE);
  }
  x->key = strdup(key);
  x->index = index;
  x->l = l;
  x->r = r;

  return x;
}

bst_t bst_init( ) {
  bst_t bst;

  bst = malloc(sizeof *bst) ;
  if (bst == NULL) {
    fprintf(stderr, "Memory allocation error.\n");
    exit(EXIT_FAILURE);
  }
  bst->N = 0;
  bst->head = ( bst->z = tree_node_new("", -1, NULL, NULL));

  return bst;
}

int bst_empty(bst_t bst) {
  return (bst==NULL || bst->N==0);
}

int bst_count(bst_t bst) {
  if (bst_empty(bst))
    return 0;

  return bst->N;
}

bst_node_t *insertR(bst_node_t *h, char *key, int i, bst_node_t *z) {
  if (h == z) {
    return tree_node_new(key, i, z, z);
  }
  if (strcmp(key, h->key) < 0) {
    h->l = insertR(h->l, key, i, z);
  } else {
    h->r = insertR(h->r, key, i, z);
  }

  return h;
}

void bst_insert_leaf(bst_t bst, char *str, int i) {
  bst->head = insertR(bst->head, str, i, bst->z);
  bst->N++;

  return;
}

int searchR(bst_node_t *h, char *key, bst_node_t *z) {
  int result;

  if (h == z) {
    return -1;
  }
  if (strcmp(key, h->key)==0) {
    return h->index;
  }
  result = searchR(h->l, key, z);
  if (result != -1) {
    return result;
  }
  result = searchR(h->r, key, z);

  return result;
}

int bst_search(bst_t bst, char *key) {
  if (bst_empty(bst)) {
    return -1;
  }
  return searchR(bst->head, key, bst->z);
}



/*
 * Copyright C.L.U.T. - Torino - Ottobre 2017
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part IV: Exam-Based Problems
 *
 */
