/*
 * Copyright C.L.U.T. - Torino - Ottobre 2017
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part IV: Exam-Based Problems
 *
 * chapter 01 - exercise 23 - solution (2-1)
 *
 * Nessuna garanzia viene fornita su questo programma:
 * gli Autori e la Casa Editrice non potranno in nessun
 * caso essere ritenuti responsabili di eventuali danni
 * (materiali e/o immateriali) da esso cagionati.
 * E' liberamente consentita la modifica (ma non la
 * redistribuzione) di questo file, a patto che le note
 * di copyright poste in testa e in coda a esso vengano
 * mantenute inalterate.
 * 
 */

#include <stdio.h>
#include <stdlib.h>

#define DX 20
#define DY 30

/* define truck size */
const int MAXX[] = {10, 10, 15, DX};
const int MAXY[] = {15, 20, 25, DY};

/* structure declaration */
typedef struct {
  int x;
  int y;
} pack_t;

/* function prototypes */
pack_t *loadPackage(int *);
void mark(int [][DY], int, int, int, int, int);
int place_r(int [][DY], int, int, pack_t *, int, int);
int check(int [][DY], int, int, int, int, int, int);
void printSolution(int [][DY], int, int);

/*
 *  main program
 */
int main(void)
{
  int truck[DX][DY], i, num, ok=0;
  pack_t *packages;

  /* load the input file and initialize the truck */
  packages = loadPackage(&num);
  mark(truck, 0, 0, DX, DY, 0);

  /* look for the solution, if any */
  for (i=0; i<4 && !ok; i++) {
    ok = place_r(truck, MAXX[i], MAXY[i], packages, num, 0);
    if (ok) {
      fprintf (stdout, "The %c-type truck must be used:\n", 'A'+i);
      printSolution(truck, MAXX[i], MAXY[i]);
    }
  } 
  if (!ok) {
    fprintf (stdout, "No available truck can be used!\n");
  }

  free(packages);          
  return EXIT_SUCCESS;
}

/* 
 *  load the package file contents
 */
pack_t *loadPackage(int *num)
{
  pack_t *packages;
  char name[20];
  FILE *fp;
  int i;

  fprintf (stdout, "Input file name: ");
  scanf("%s", name);

  fp = fopen(name, "r");
  if (fp==NULL) {
    fprintf (stderr, "File open error.\n");
    exit(EXIT_FAILURE);
  }

  fscanf(fp, "%d", num);
  packages= (pack_t *)malloc((*num) * sizeof(pack_t));
  if (packages == NULL) {
    fprintf (stderr, "Memory allocation error.\n");
    exit(EXIT_FAILURE);
  }

  for (i=0; i<*num; i++) {
    fscanf(fp, "%dx%d", &packages[i].x, &packages[i].y);
  }

  fclose(fp);

  return packages;
}

/* 
 * mark some positions of the truck with a given id
 */
void mark(int truck[][DY], int x, int y, int dimX, int dimY, int id)
{
  int i, j;

  for (i=x; i<x+dimX; i++) {
    for (j=y; j<y+dimY; j++) {
      truck[i][j] = id;
    }
  }
}

/* 
 *  place the set of packages into the truck, recursive function
 */
int place_r(int truck[][DY], int maxX, int maxY, pack_t *packages, int num, int id)
{
  int i, j, x, y, m;

  if (id == num) {
    return 1;
  }

  x = packages[id].x;
  y = packages[id].y;
  m = (x<y) ? x : y;
  for (i=0; i<=maxX-m; i++) {
    for (j=0; j<=maxY-m; j++) {
      if (check(truck, maxX, maxY, i, j, x, y)) {
        mark(truck, i, j, x, y, id+1);
        if (place_r(truck, maxX, maxY, packages, num, id+1)) {
          return 1;
        }
        mark(truck, i, j, x, y, 0);
      }

      if (check(truck, maxX, maxY, i, j, y, x)) {
        mark(truck, i, j, y, x, id+1);
        if (place_r(truck, maxX, maxY, packages, num, id+1)) {
          return 1;
        }
        mark(truck, i, j, y, x, 0);
      }
    }
  }

  return 0;
}

/* 
 * check whether a given package can be placed in a given position
 */
int check(int truck[][DY], int maxX, int maxY, int x, int y, int dimX, int dimY)
{
  int i, j;

  if (x+dimX>maxX || y+dimY>maxY) {
    return 0;
  }

  for (i=x; i<x+dimX; i++) {
    for (j=y; j<y+dimY; j++) {
      if (truck[i][j] != 0) {
        return 0;
      }
    }
  }

  return 1;
}

/* 
 *  print the final disposition on screen
 */
void printSolution(int truck[][DY], int maxX, int maxY)
{
  int i, j;

  for (i=0; i<maxX; i++) {
    for (j=0; j<maxY; j++) {
      if (truck[i][j] == 0) {
        /* just for a better printing */
        fprintf (stdout, ".");
      } else {
        fprintf (stdout, "%d", truck[i][j]);
      }
    }
    fprintf (stdout, "\n");
  }
}

/*
 * Copyright C.L.U.T. - Torino - Ottobre 2017
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part IV: Exam-Based Problems
 *
 */
