/*
 * Copyright C.L.U.T. - Torino - Ottobre 2017
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part IV: Exam-Based Problems
 *
 * chapter 01 - exercise 27 - solution (2-1)
 *
 * Nessuna garanzia viene fornita su questo programma:
 * gli Autori e la Casa Editrice non potranno in nessun
 * caso essere ritenuti responsabili di eventuali danni
 * (materiali e/o immateriali) da esso cagionati.
 * E' liberamente consentita la modifica (ma non la
 * redistribuzione) di questo file, a patto che le note
 * di copyright poste in testa e in coda a esso vengano
 * mantenute inalterate.
 * 
 */

#include <stdlib.h>
#include <string.h>
#include "symbolTable.h"

typedef struct st_item_s {
  char *id;
} st_item_t;

struct st_s {
  st_item_t *items;
  int N, M;
};

st_t st_init(int maxN) {
    st_t t;
    t = calloc(1, sizeof(*t));
    t->N = 0;
    t->M = maxN;
    t->items = calloc(t->M, sizeof(st_item_t));
    return t;
}

int st_get_size(st_t t) {
  return t->N;
}

int st_add(st_t t, char *str) {
  if (t->N == t->M) {
    t->M *= 2;
    t->items = realloc(t->items, t->M*sizeof(st_item_t));
  }
  t->items[t->N++].id = strdup(str);
  return t->N-1;
}

int st_find(st_t t, char *str) {
  int i;
  for(i=0;i<t->N;i++)
    if (!strcmp(t->items[i].id, str)) return i;
  return -1;
}

int st_find_or_add(st_t t, char *str) {
  int id = st_find(t, str);
  if (id == -1) {
    return st_add(t, str);
  }
  return id;
}

void st_dispose(st_t t) {
  int i;
  if (t == NULL) return;
  if (t->items != NULL) {
    for(i=0;i<t->N;i++) {
      if (t->items[i].id != NULL) free(t->items[i].id);
    }
    free(t->items);
  }
  free(t);
}

char *st_read(st_t t, int index) {
  if (index < 0 || index > t->N) return NULL;
  return t->items[index].id;
}

/*
 * Copyright C.L.U.T. - Torino - Ottobre 2017
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part IV: Exam-Based Problems
 *
 */
