/*
 * Copyright C.L.U.T. - Torino - Ottobre 2017
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part IV: Exam-Based Problems
 *
 * chapter 01 - exercise 29 - solution (2-1)
 *
 * Nessuna garanzia viene fornita su questo programma:
 * gli Autori e la Casa Editrice non potranno in nessun
 * caso essere ritenuti responsabili di eventuali danni
 * (materiali e/o immateriali) da esso cagionati.
 * E' liberamente consentita la modifica (ma non la
 * redistribuzione) di questo file, a patto che le note
 * di copyright poste in testa e in coda a esso vengano
 * mantenute inalterate.
 * 
 */

#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include <stdio.h>

#define MAXS 21
#define PLUS_INF 1e6

typedef struct node_s {
  int off;
  struct node_s *next;
} node_t;

// A single product
typedef struct product_s {
  char name[MAXS];
  int price;
  node_t *offer_list; 
} product_t;

// A single offer
typedef struct offer_s {
  char name[MAXS];
  int price;
  int *pro;
  int *qta;
  int n;
} offer_t;

// Array of products
typedef struct products_s {
  product_t *array;
  int n;
} products_t;

// Array of offers
typedef struct offers_s {
  offer_t *array;
  int n;
} offers_t;

// Prototypes
void product_read(char *, products_t *, offers_t *);
int product_find(products_t, char *);
void offer_add(product_t *, int);
node_t *offer_init(int, node_t *);
void offer_per_product_write(char *, products_t, offers_t);
void offer_name_write(offers_t, int);
void common_product_write(char *, char *, products_t, offers_t);
int offer_find(offers_t, char *);
void product_name_write(products_t, int);
int *list_read(char *, products_t);
void mis_cost_grocery(products_t, offers_t, int *);
void powerset(int, int *, int *, int *, int, products_t, offers_t, int *);
void check (int *, int *, int *, int, products_t, offers_t, int *);

int main(int argc, char **argv) {
  products_t P;
  offers_t O;
  int *grocery;

  fprintf (stdout, "Reading ...\n");
  product_read (argv[1], &P, &O);
  
  // Problem 1: Bargains with same product
  fprintf (stdout, "Offer per product ...\n");
  offer_per_product_write ("Nutella", P, O);

  // Problem 2: Given 2 bargains find common products
  fprintf (stdout, "Common products ...\n");
  common_product_write ("OFF1", "OFF4", P, O);

  // Problem 3: Find minimum cost
  fprintf (stdout, "Best grocery ...\n");
  grocery = list_read (argv[2], P);
  mis_cost_grocery(P, O, grocery);
  
  return 0;
}

/*
 *  Read all products and all bargain offers
 */
void product_read(char *name, products_t *P, offers_t *O) {
  FILE *fp;
  int i, j, index;
  char str[MAXS];

  fp = fopen (name, "r");
  if (fp == NULL) {
    fprintf (stderr, "File open error.\n");
    exit(EXIT_FAILURE);
  }

  fscanf(fp, "%d %d", &P->n, &O->n);
  P->array = calloc(P->n, sizeof(product_t));
  if (P->array==NULL) {
    fprintf (stderr, "Memory allocation error.\n");
    exit(EXIT_FAILURE);
  }

  for(i=0;i<P->n;i++) {
    fscanf(fp, "%s %d",
      P->array[i].name, &P->array[i].price);
  }

  O->array = calloc(O->n, sizeof(offer_t));
  if (O->array==NULL) {
    fprintf (stderr, "Memory allocation error.\n");
    exit(EXIT_FAILURE);
  }
  for(i=0;i<O->n;i++) {
    fscanf(fp, "%s %d %d",
      O->array[i].name, &O->array[i].price, &O->array[i].n);
    O->array[i].pro = calloc(O->array[i].n, sizeof(int));
    O->array[i].qta = calloc(O->array[i].n, sizeof(int));
    if (O->array[i].pro==NULL || O->array[i].qta) {
      fprintf (stderr, "Memory allocation error.\n");
      exit(EXIT_FAILURE);
    }
    for (j=0; j<O->array[i].n; j++) {
      fscanf(fp, "%s %d", str, &O->array[i].qta[j]);
      index = product_find(*P, str);
      assert(index != -1);
      O->array[i].pro[j] = index;
      offer_add(&(P->array[index]), j);
    }
  }

  fclose(fp);

  return;
}

// Find a product within the array of products
int product_find(products_t P, char *name) {
  int i;

  for(i=0;i<P.n;i++) {
    if (!strcmp(P.array[i].name, name))
      return i;
  }
  return -1;
}

// Add a bargain offer reference to a product
void offer_add(product_t *p, int off) {
  p->offer_list = offer_init(off, p->offer_list);

  return;
}

node_t *offer_init(int off, node_t *next) {
  node_t *n;

  n = calloc(1, sizeof(node_t));
  if (n==NULL) {
    fprintf (stderr, "Memory allocation error.\n");
    exit(EXIT_FAILURE);
  }

  n->off = off;
  n->next = next;

  return n;
}

/*
 *  Given a product find all bargain offers including it
 */
void offer_per_product_write(char *name, products_t P, offers_t O) {
  int p = product_find(P, name);
  if (p == -1) return;

  node_t *iter = P.array[p].offer_list;
  while(iter != NULL) {
    offer_name_write(O, iter->off);
    iter = iter->next;
  }
}

void offer_name_write(offers_t O, int off) {
  if (O.n <= off || off < 0)
    return;

  fprintf (stdout, "%s\n", O.array[off].name); 
}

/*
 *  Given two bargain offers find common products
 */
void common_product_write(char *name1, char *name2, products_t P, offers_t O) {
  int o1 = offer_find(O, name1);
  int o2 = offer_find(O, name2), i;
  int *mark;

  if (o1 == -1 || 02 == -1) {
    return;
  }

  mark = calloc(P.n, sizeof(int));
  if (mark==NULL) {
    fprintf (stderr, "Memory allocation error.\n");
    exit(EXIT_FAILURE);
  }

  for (i=0; i<O.array[o1].n; i++) {
    mark[O.array[o1].pro[i]]++;
  } 
  for (i=0; i<O.array[o2].n; i++) {
    mark[O.array[o2].pro[i]]++; 
  } 
  for (i=0; i<P.n; i++) {
    if(mark[i] == 2) {
      product_name_write(P, i);
    }
  }

  free(mark);
}

// Find a bargain offer within the array of bargain offers
int offer_find(offers_t O, char *name) {
  int i;

  for(i=0; i<O.n; i++) {
    if (!strcmp(O.array[i].name, name))
      return i;
  }

  return -1;
}

// Print-out a product name
void product_name_write(products_t P, int prod) {
  if (P.n <= prod || prod < 0)
    return;

  fprintf (stdout, "%s\n", P.array[prod].name); 
}

/*
 *  Read the grocery list from file
 */
int *list_read(char *name, products_t P) {
  FILE *fp;
  int *grocery, i, qta, id, ns;
  char str[MAXS];

  grocery = calloc(P.n, sizeof(int)), i, qta, id, ns;
  if (grocery==NULL) {
    fprintf (stderr, "Memory allocation error.\n");
    exit(EXIT_FAILURE);
  }

  fp = fopen (name, "r");
  if (fp == NULL) {
    fprintf (stderr, "File open error.\n");
    exit(EXIT_FAILURE);
  }

  fscanf(fp, "%d", &ns);
  for (i=0; i<ns; i++) {
    fscanf(fp, "%s %d", str, &qta);
    id = product_find(P, str);
    if (id != -1)
      grocery[id] = qta;
  }

  fclose(fp);

  return grocery;
}

/*
 *  Find the optimum set of purchases
 */
void mis_cost_grocery(products_t P, offers_t O, int *grocery) {
  int *sol = calloc(O.n, sizeof(int));
  int *best_sol = calloc(O.n, sizeof(int));
  int bestprice = PLUS_INF;

  if (sol==NULL || best_sol==NULL) {
    fprintf (stderr, "Memory allocation error.\n");
    exit(EXIT_FAILURE);
  }

  powerset(0, sol, best_sol, &bestprice, 0, P, O, grocery); 

  fprintf (stdout, "Total cost: %d\n", bestprice);
  
  free(sol);
  free(best_sol);
}

// Generate the powerset
void powerset(
  int pos, int *sol, int *best_sol, int *bestp,
  int currp, products_t P, offers_t O, int *grocery
) {
  if (pos == O.n) {
    check(sol, best_sol, bestp, currp, P, O, grocery);
    return;  
  }
  if ((currp + O.array[pos].price) <= *bestp) {
    sol[pos] = 1;
    powerset(pos+1, sol, best_sol, bestp, currp + O.array[pos].price,
      P, O, grocery);
  }
  sol[pos] = 0;
  powerset(pos+1, sol, best_sol, bestp, currp, P, O, grocery);

  return;
}

// Check current grocery list
void check (
  int *sol, int *best_sol, int *bestp, int currp, products_t P,
  offers_t O, int *grocery
) {
  int i, j, *current_grocery;

  current_grocery = calloc(P.n, sizeof(int));
  if (current_grocery==NULL) {
    fprintf (stderr, "Memory allocation error.\n");
    exit(EXIT_FAILURE);
  }

  // Create current grocery list
  for (i=0; i<O.n; i++) {
    if (!sol[i]) {
      continue;
    }
    for (j=0; j<O.array[i].n; j++) {
      current_grocery[O.array[i].pro[j]] += O.array[i].qta[j];
    }
  }

  // Find missing items
  for (i=0; i<P.n; i++) {
    if (current_grocery[i] < grocery[i]) {
      currp += ((grocery[i]-current_grocery[i]) * P.array[i].price);
    }
  }

  if (currp < *bestp) {
    *bestp = currp;
    for (i=0; i<O.n; i++) {
      best_sol[i] = sol[i];
    }
  }

  free(current_grocery);

  return;
}




/*
 * Copyright C.L.U.T. - Torino - Ottobre 2017
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part IV: Exam-Based Problems
 *
 */
