/*
 * Copyright C.L.U.T. - Torino - Ottobre 2017
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part IV: Exam-Based Problems
 *
 * chapter 01 - exercise 31 - solution (2-1)
 *
 * Nessuna garanzia viene fornita su questo programma:
 * gli Autori e la Casa Editrice non potranno in nessun
 * caso essere ritenuti responsabili di eventuali danni
 * (materiali e/o immateriali) da esso cagionati.
 * E' liberamente consentita la modifica (ma non la
 * redistribuzione) di questo file, a patto che le note
 * di copyright poste in testa e in coda a esso vengano
 * mantenute inalterate.
 * 
 */

#include <stdlib.h>
#include <string.h>
#include <float.h>
#include <stdio.h>

#define MAX_LEN  51
// Set DEBUG to TRUE to have some extra debugging feature
#define DEBUG     1

typedef struct client_s {
  char name[MAX_LEN];
  int in;
  int out;
  int allocation_current;
  int allocation_best;
} client_t;

void file_read (char *, int *, int *, int **, int *, int *, int *, client_t **);
#if DEBUG
void struct_write (int, int, int *, int, int, int, client_t *);
#endif
void solve_r (int, int, int *, int, int, int, client_t *, int, float *);
void check (int, int, int *, int, int, int, client_t *, float *);

int main(int argc, char **argv) {
  int D, N, A, B, *room, reservation;
  float cost;
  client_t *client;

  file_read (argv[1], &D, &N, &room, &A, &B, &reservation, &client);
#if DEBUG
  struct_write (D, N, room, A, B, reservation, client);
#endif

  cost = (-1.0);
  solve_r (D, N, room, A, B, reservation, client, 0, &cost);
  if (cost > 0.0) {
    fprintf (stdout, "Solution found.\n");
    struct_write (D, N, room, A, B, reservation, client);
  }

  free (room);
  free (client);

  return 0;
}

/*
 *  Parse input file and fill-in array of structures client_t 
 */
void file_read (
  char *name, int *D, int *N, int **room, int *A, int *B,
  int *reservation, client_t **client
) {
  FILE *fp;
  char str[MAX_LEN];
  int i, *tmp1;
  client_t *tmp2;
  
  fp = fopen (name, "r");
  if (fp == NULL) {
    fprintf (stderr, "File open error.\n");
    exit(EXIT_FAILURE);
  }
  
  if (fscanf(fp, "D=%d N=%d", D, N) != 2) {
    fprintf (stderr, "Error reading file 1.\n");
    exit(EXIT_FAILURE);
  }

  tmp1 = malloc ((*N) * sizeof (int));
  if (tmp1 == NULL) {
    fprintf (stderr, "Memory allocation error.\n");
    exit(EXIT_FAILURE);
  }
  for (i=0; i<*N; i++) {
    if (fscanf(fp, "%s", str) != 1) {
      fprintf (stderr, "Error reading file 2.\n");
      exit(EXIT_FAILURE);
    }
    sscanf (str, "R%*d=%d", &tmp1[i]);
  }

  *room = tmp1;

  if (fscanf(fp, "%s", str) != 1) {
    fprintf (stderr, "Error reading file 3.\n");
    exit(EXIT_FAILURE);
  }
  sscanf (str, "A=%d", A);

  if (fscanf(fp, "%s", str) != 1) {
    fprintf (stderr, "Error reading file 3.\n");
    exit(EXIT_FAILURE);
  }
  sscanf (str, "B=%d", B);

  if (fscanf(fp, "%s", str) != 1) {
    fprintf (stderr, "Error reading file 4.\n");
    exit(EXIT_FAILURE);
  }
  sscanf (str, "#reservations=%d", reservation);
  
  tmp2 = malloc ((*reservation) * sizeof (client_t));
  if (tmp2 == NULL) {
    fprintf (stderr, "Memory allocation error.\n");
    exit(EXIT_FAILURE);
  }
  for (i=0; i<*reservation; i++) {
    if (fscanf(fp, "%s", str) != 1) {
      fprintf (stderr, "Error reading file 5.\n");
      exit(EXIT_FAILURE);
    }
    sscanf (str, "client=%s", tmp2[i].name);
    if (fscanf(fp, "%s", str) != 1) {
      fprintf (stderr, "Error reading file 5.\n");
      exit(EXIT_FAILURE);
    }
    sscanf (str, "in=%d", &tmp2[i].in);
    if (fscanf(fp, "%s", str) != 1) {
      fprintf (stderr, "Error reading file 5.\n");
      exit(EXIT_FAILURE);
    }
    sscanf (str, "out=%d", &tmp2[i].out);
    tmp2[i].allocation_best = (-1);
    tmp2[i].allocation_current = (-1);
  }
  *client = tmp2;

  fclose (fp);

  return;
}

/*
 *  Allocate each customer in a random room
 */
void solve_r (
  int D, int N, int *room, int A, int B, int reservation, client_t *client,
  int level, float *cost
) {
  int i;
#if DEBUG
  static int xxx = 0;
#endif

  if (level >= reservation) {
#if DEBUG
    fprintf (stdout, "#### SOLUTION %d ###", ++xxx);
    struct_write (D, N, room, A, B, reservation, client);
#endif
    check (D, N, room, A, B, reservation, client, cost);
    return;
  }

  for (i=0; i<N; i++) {
    client[level].allocation_current = i;
    solve_r (D, N, room, A, B, reservation, client, level+1, cost);
    client[level].allocation_current = (-1);
  }
}

/*
 *  Verify constraints
 */
void check (
  int D, int N, int *room, int A, int B, int reservation,
  client_t *client, float *cost
) {
  int **mat, i, j, ok;
  float *h, h_min, h_max, h_sum, cost_current;

  // Check max number of reservations
  mat = malloc (D * sizeof (int *));
  if (mat == NULL) {
    fprintf (stderr, "Memory allocation error.\n");
    exit(EXIT_FAILURE);
  }
  for (i=0; i<D; i++) {
    mat[i] = malloc (N * sizeof (int));
    if (mat[i] == NULL) {
      fprintf (stderr, "Memory allocation error.\n");
      exit(EXIT_FAILURE);
    }
  }
  for (i=0; i<D; i++) {
    for (j=0; j<N; j++) {
      mat[i][j] = 0;
    }
  }
  for (i=0; i<reservation; i++) {
    for (j=client[i].in; j<client[i].out; j++) {
      mat[j][client[i].allocation_current]++;
    }
  }
#if DEBUG
  fprintf (stdout, "MAT\n");
  for (i=0; i<D; i++) {
    for (j=0; j<N; j++) {
      fprintf (stdout, "%d ", mat[i][j]);
    }
    fprintf (stdout, "\n");
  }
#endif
  ok = 1;
  for (i=0; i<D && ok; i++) {
    for (j=0; j<N && ok; j++) {
      if (mat[i][j] > room[j]) {
        ok = 0;
      }
    }
  }

  // IFF number of reservation is OK compute cost function
  if (ok!=0) {
#if DEBUG
    fprintf (stdout, "Reservations OK\n");
#endif

    h = malloc (N * sizeof (float));
    if (h == NULL) {
      fprintf (stderr, "Memory allocation error.\n");
      exit(EXIT_FAILURE);
    }
    for (i=0; i<N; i++) {
      h[i] = 0.0;
    }
    // Compute all h_i
    for (i=0; i<D; i++) {
      for (j=0; j<N; j++) {
        h[j] += mat[i][j];
      }
    }
    fprintf (stdout, "H1: ");
    for (j=0; j<N; j++) {
      fprintf (stdout, "%6.2f ", h[j]);
    }
    for (j=0; j<N; j++) {
      h[j] = h[j] / (room[j] * D);
    }
    fprintf (stdout, "\nH2: ");
    for (j=0; j<N; j++) {
      fprintf (stdout, "%6.2f ", h[j]);
    }
    fprintf (stdout, "\n");
    h_min = h_max = (-1);
    h_sum= 0;
    for (j=0; j<N; j++) {
      h_sum += h[j];
      if (h_min==(-1) || h[j]<h_min) {
        h_min = h[j];
      }
      if (h_max==(-1) || h[j]>h_max) {
        h_max = h[j];
      }
    }
    cost_current = A * h_sum - B * (h_max - h_min);
    fprintf (stdout, "h_min=%6.2f h_max=%6.2f h_sum=%6.2f cost=%6.2f\n",
      h_min, h_max, h_sum, cost_current);
    if (cost_current>(*cost)) {
      *cost = cost_current;
      for (i=0; i<reservation; i++) {
        client[i].allocation_best = client[i].allocation_current;
      }
    }

    free (h);
  }

  for (i=0; i<D; i++) {
    free (mat[i]);
  }
  free (mat);

  return;
}

#if DEBUG
void struct_write (
  int D, int N, int *room, int A, int B, int reservation, client_t *client
) {
  int i;
  
  fprintf (stdout, "D=%d N=%d ", D, N);
  for (i=0; i<N; i++) {
    fprintf (stdout, "R%d=%d ", i+1, room[i]);
  }
  fprintf (stdout, "\n");
  fprintf (stdout, "A=%d B=%d\n", A, B);
  fprintf (stdout, "#reservations=%d\n", reservation);
  for (i=0; i<reservation; i++) {
    fprintf (stdout, "client=%s in=%d out=%d allocCurrent=%d allocBest=%d\n",
      client[i].name, client[i].in, client[i].out,
      client[i].allocation_current, client[i].allocation_best);
  }
  
  return;
}
#endif

/*
 * Copyright C.L.U.T. - Torino - Ottobre 2017
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part IV: Exam-Based Problems
 *
 */
