/*
 * Copyright C.L.U.T. - Torino - Ottobre 2017
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part IV: Exam-Based Problems
 *
 * chapter 01 - exercise 33 - solution (2-1)
 *
 * Nessuna garanzia viene fornita su questo programma:
 * gli Autori e la Casa Editrice non potranno in nessun
 * caso essere ritenuti responsabili di eventuali danni
 * (materiali e/o immateriali) da esso cagionati.
 * E' liberamente consentita la modifica (ma non la
 * redistribuzione) di questo file, a patto che le note
 * di copyright poste in testa e in coda a esso vengano
 * mantenute inalterate.
 * 
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#define NC 51
// Set to TRUE to debug
#define DEBUG 0

// A bank account
typedef struct bank_s {
  char bank[NC];
  char branch[NC];
  char abi[NC];
  char cab[NC];
  char iban[NC];
  int deposit;
  struct bank_s *next;
} bank_t;

// A bank customer
typedef struct customer_s {
  char name1[NC];
  char name2[NC];
  struct bank_s *bank_list;
  struct customer_s *left;
  struct customer_s *right;
} customer_t;

// Prototypes
void find_and_print (customer_t *, char *, char *);
void file1_read (customer_t **, FILE *);
void file2_read (char *, char *, char *, char*, FILE *, customer_t **);
void tree_insert (customer_t **,customer_t*,bank_t*);
void tree_visit (customer_t *);
void tree_dispose (customer_t *);

int main (int argc, char *argv[]) {
  char name1[NC], name2[NC];;
  customer_t *tree;
  FILE *fp1;

  tree=NULL;
  if (argc<=1 || argc>2) {
    fprintf (stderr, "Run as: pgrm_name file_name\n");
    return (1);
  }

  fp1=fopen(argv[1],"r");
  if (fp1==NULL) {
    fprintf (stderr, "File open error (%s).\n", argv[1]);
    exit(EXIT_FAILURE);
  }
  file1_read(&tree,fp1);
  fclose(fp1);

#if DEBUG
  tree_visit(tree);
#endif

  do { 
    fprintf (stdout, "Surnname+name ('exit exit' to end): ");
    scanf("%s %s", name1, name2);
    if (strcmp(name1,"exit")==0 && strcmp(name2,"exit")==0) {
      fprintf (stdout, "Bye bye.\n");
      return (0);
    }

    find_and_print (tree, name1, name2);

  } while (strcmp(name1,"exit")!=0);

  tree_dispose (tree);

  return (0);
}

/*
 *  Recursive BST search and list visit
 */
void find_and_print (customer_t *root, char *name1, char *name2) {
  customer_t *app;
  bank_t *app1;

  app = root;
  if(root==NULL) {
    fprintf (stderr, "No data\n");
  } else {
    if (strcmp(app->name1, name1)!=0 || strcmp(app->name2,name2)!=0) {
      if(strcmp(app->name1, name1)<0) {
         find_and_print(app->right,name1,name2);
      } else {
        if(strcmp(app->name1, name1)>0)  {
           find_and_print(app->left,name1,name2);
        } else {
          if (strcmp(app->name1, name1)==0) {
            if(strcmp(app->name2, name2)<0) {
              find_and_print(app->right,name1,name2);
            } else {
              find_and_print(app->left,name1,name2);
            }
          }
        }
      }
    } else {
      app1=app->bank_list;
      while(app1!=NULL){
        fprintf (stdout,
          "bank: %s -- branch: %s -- abi: %s -- cab: %s -- iban: %s -- deposit: %d\n",
          app1->bank,app1->branch,app1->abi,
        app1->cab,app1->iban,app1->deposit);
        app1=app1->next;
      }
    }
  }

  return;   
}

/*
 *  Read "bank account" file
 */
void file1_read (customer_t **root, FILE *fp) {
  char bank[NC], branch[NC], abi[NC], cab[NC], name2[NC];
  FILE *fp2;

  while (fscanf(fp,"%s %s %s %s %s", bank, branch, abi, cab, name2)!=EOF) {

    fp2 = fopen(name2,"r");
    if (fp2==NULL) {
      fprintf (stderr, "File open error (%s).\n", name2);
      exit(EXIT_FAILURE);
    }

    file2_read(bank,branch,abi,cab,fp2,root);

    fclose(fp2);
  }

  return;
}

/*
 *  Read "customer" file (for each bank account)
 */
void file2_read (
  char *bank, char *branch, char *abi, char *cab, FILE *fp, customer_t **root
) {
  char name1[NC], name2[NC], iban[NC];
  int deposit;
  customer_t *itemc;
  bank_t *item;

  while(fscanf(fp,"%s %s %s %d", name1, name2, iban, &deposit) != EOF) {
    item = (bank_t *) calloc (1,sizeof(bank_t));
    itemc = (customer_t *) calloc (1,sizeof(customer_t)); 
    if (item==NULL || itemc==NULL) {
      fprintf (stderr, "Memory allocation error.\n");
      exit(EXIT_FAILURE);
    }

    strcpy(itemc->name1, name1);
    strcpy(itemc->name2, name2);
    strcpy(item->iban, iban);
    item->deposit = deposit;
    strcpy(item->bank,bank);
    strcpy(item->branch, branch);
    strcpy(item->abi,abi);
    strcpy(item->cab,cab);               

    tree_insert(root, itemc, item);
  }

  return;
}

/*
 *  Modified library function: Recursive BST insertion
 */
void tree_insert (customer_t **root, customer_t *item, bank_t *bank_list) {
  customer_t *app;

  app = *root;

  if(app==NULL){
    item->bank_list = bank_list;
    *root = item;
  } else{
    if(strcmp(item->name1,app->name1)<0) {
      tree_insert(&app->left,item, bank_list);
    } else {
      if(strcmp(item->name1,app->name1)==0 && strcmp(item->name2,app->name2)<0) {
        tree_insert(&app->left,item, bank_list);
      } else {
        if (strcmp(item->name1,app->name1)>0) {
          tree_insert(&app->right, item, bank_list);
	} else {
          if(strcmp(item->name1,app->name1)==0 && strcmp(item->name2,app->name2)>0) {
            tree_insert(&app->right, item, bank_list);  
          } else {
            if (strcmp(item->name1,app->name1)==0 &&
                strcmp(item->name2,app->name2)==0){
              bank_list->next = app->bank_list;
              app->bank_list = bank_list;
              free (item);
            }
	  }
        }
      }
    }
  }

  return;         
}

/*
 *  Modified library function: Recursive BST visit
 */
void tree_visit (customer_t *tree) {
  customer_t *app;
  bank_t *app1;

  app = tree;
  if (app!=NULL) {
    tree_visit(app->left);
    fprintf (stdout, "%s %s\n", app->name1, app->name2);
    app1 = app->bank_list;

    while(app1!=NULL) {
      fprintf (stdout, "%s %s %s %s %s %d\n",
        app1->bank, app1->branch, app1->abi, app1->cab, app1->iban, app1->deposit);
      app1=app1->next;
    }
    tree_visit(app->right);
  }

  return;
}

/*
 *  Free tree and lists
 */
void tree_dispose (customer_t *tree) {
  bank_t *p1, *p2;

  if (tree == NULL) {
    return;
  }

  tree_visit(tree->left);
  tree_visit(tree->right);

  p1 = tree->bank_list;
  while(p1!=NULL) {
    p2 = p1->next;
    free (p1);
    p1 = p2;
  }

  free (tree);

  return;
}

/*
 * Copyright C.L.U.T. - Torino - Ottobre 2017
 *
 * Stefano Quer
 * Advanced Programming and Problem-Solving Strategies in C
 * Part IV: Exam-Based Problems
 *
 */
